package java.util.stream;

import java.util.Objects;
import java.util.function.IntFunction;
import java.util.function.Supplier;
import java.util.Spliterator;

abstract class AbstractPipeline<E_IN, E_OUT, S extends BaseStream<E_OUT, S>> extends PipelineHelper<E_OUT> implements BaseStream<E_OUT, S>
{
    private static final String MSG_STREAM_LINKED = "stream has already been operated upon or closed";
    private static final String MSG_CONSUMED = "source already consumed or closed";
    private final AbstractPipeline sourceStage;
    private final AbstractPipeline previousStage;
    protected final int sourceOrOpFlags;
    private AbstractPipeline nextStage;
    private int depth;
    private int combinedFlags;
    private Spliterator<?> sourceSpliterator;
    private Supplier<? extends Spliterator<?>> sourceSupplier;
    private boolean linkedOrConsumed;
    private boolean sourceAnyStateful;
    private Runnable sourceCloseAction;
    private boolean parallel;
    
    AbstractPipeline(Supplier<? extends Spliterator<?>> sourceSupplier, int n, boolean parallel) {
        this.previousStage = null;
        this.sourceSupplier = sourceSupplier;
        this.sourceStage = this;
        this.sourceOrOpFlags = (n & StreamOpFlag.STREAM_MASK);
        this.combinedFlags = (~(this.sourceOrOpFlags << 1) & StreamOpFlag.INITIAL_OPS_VALUE);
        this.depth = 0;
        this.parallel = parallel;
    }
    
    AbstractPipeline(Spliterator<?> sourceSpliterator, int n, boolean parallel) {
        this.previousStage = null;
        this.sourceSpliterator = sourceSpliterator;
        this.sourceStage = this;
        this.sourceOrOpFlags = (n & StreamOpFlag.STREAM_MASK);
        this.combinedFlags = (~(this.sourceOrOpFlags << 1) & StreamOpFlag.INITIAL_OPS_VALUE);
        this.depth = 0;
        this.parallel = parallel;
    }
    
    AbstractPipeline(AbstractPipeline<?, E_IN, ?> previousStage, int n) {
        if (previousStage.linkedOrConsumed) {
            throw new IllegalStateException("stream has already been operated upon or closed");
        }
        previousStage.linkedOrConsumed = true;
        previousStage.nextStage = this;
        this.previousStage = previousStage;
        this.sourceOrOpFlags = (n & StreamOpFlag.OP_MASK);
        this.combinedFlags = StreamOpFlag.combineOpFlags(n, previousStage.combinedFlags);
        this.sourceStage = previousStage.sourceStage;
        if (this.opIsStateful()) {
            this.sourceStage.sourceAnyStateful = true;
        }
        this.depth = previousStage.depth + 1;
        return;
    }
    
    final <R> R evaluate(TerminalOp<E_OUT, R> terminalOp) {
        assert this.getOutputShape() == terminalOp.inputShape();
        if (this.linkedOrConsumed) {
            throw new IllegalStateException("stream has already been operated upon or closed");
        }
        this.linkedOrConsumed = true;
        return this.isParallel() ? terminalOp.evaluateParallel(this, this.sourceSpliterator(terminalOp.getOpFlags())) : terminalOp.evaluateSequential(this, this.sourceSpliterator(terminalOp.getOpFlags()));
    }
    
    final Node<E_OUT> evaluateToArrayNode(IntFunction<E_OUT[]> intFunction) {
        if (this.linkedOrConsumed) {
            throw new IllegalStateException("stream has already been operated upon or closed");
        }
        this.linkedOrConsumed = true;
        if (this.isParallel() && this.previousStage != null && this.opIsStateful()) {
            this.depth = 0;
            return this.opEvaluateParallel(this.previousStage, this.previousStage.sourceSpliterator(0), intFunction);
        }
        return this.evaluate(this.sourceSpliterator(0), true, intFunction);
    }
    
    final Spliterator<E_OUT> sourceStageSpliterator() {
        if (this != this.sourceStage) {
            throw new IllegalStateException();
        }
        if (this.linkedOrConsumed) {
            throw new IllegalStateException("stream has already been operated upon or closed");
        }
        this.linkedOrConsumed = true;
        if (this.sourceStage.sourceSpliterator != null) {
            Spliterator<Object> sourceSpliterator = this.sourceStage.sourceSpliterator;
            this.sourceStage.sourceSpliterator = null;
            return sourceSpliterator;
        }
        if (this.sourceStage.sourceSupplier != null) {
            Spliterator<Object> spliterator = (Spliterator)this.sourceStage.sourceSupplier.get();
            this.sourceStage.sourceSupplier = null;
            return spliterator;
        }
        throw new IllegalStateException("source already consumed or closed");
    }
    
    @Override
    public final S sequential() {
        this.sourceStage.parallel = false;
        return this;
    }
    
    @Override
    public final S parallel() {
        this.sourceStage.parallel = true;
        return this;
    }
    
    @Override
    public void close() {
        this.linkedOrConsumed = true;
        this.sourceSupplier = null;
        this.sourceSpliterator = null;
        if (this.sourceStage.sourceCloseAction != null) {
            Runnable sourceCloseAction = this.sourceStage.sourceCloseAction;
            this.sourceStage.sourceCloseAction = null;
            sourceCloseAction.run();
        }
    }
    
    @Override
    public S onClose(Runnable obj) {
        Objects.requireNonNull(obj);
        Runnable sourceCloseAction = this.sourceStage.sourceCloseAction;
        this.sourceStage.sourceCloseAction = ((sourceCloseAction == null) ? obj : Streams.composeWithExceptions(sourceCloseAction, obj));
        return this;
    }
    
    @Override
    public Spliterator<E_OUT> spliterator() {
        if (this.linkedOrConsumed) {
            throw new IllegalStateException("stream has already been operated upon or closed");
        }
        this.linkedOrConsumed = true;
        if (this == this.sourceStage) {
            if (this.sourceStage.sourceSpliterator != null) {
                Spliterator<Object> sourceSpliterator = this.sourceStage.sourceSpliterator;
                this.sourceStage.sourceSpliterator = null;
                return sourceSpliterator;
            }
            if (this.sourceStage.sourceSupplier != null) {
                Supplier<Object> sourceSupplier = this.sourceStage.sourceSupplier;
                this.sourceStage.sourceSupplier = null;
                return this.lazySpliterator(sourceSupplier);
            }
            throw new IllegalStateException("source already consumed or closed");
        }
        else {
            return this.wrap(this, ()/*  */ -> this.sourceSpliterator(0), this.isParallel());
        }
    }
    
    @Override
    public final boolean isParallel() {
        return this.sourceStage.parallel;
    }
    
    final int getStreamFlags() {
        return StreamOpFlag.toStreamFlags(this.combinedFlags);
    }
    
    private Spliterator<?> sourceSpliterator(int n) {
        Spliterator<Object> spliterator;
        if (this.sourceStage.sourceSpliterator != null) {
            spliterator = this.sourceStage.sourceSpliterator;
            this.sourceStage.sourceSpliterator = null;
        }
        else if (this.sourceStage.sourceSupplier != null) {
            spliterator = (Spliterator)this.sourceStage.sourceSupplier.get();
            this.sourceStage.sourceSupplier = null;
        }
        else {
            throw new IllegalStateException("source already consumed or closed");
        }
        if (this.isParallel() && this.sourceStage.sourceAnyStateful) {
            int n2 = 1;
            AbstractPipeline<Object, Object, Object> sourceStage = this.sourceStage;
            AbstractPipeline<Object, Object, Object> abstractPipeline = this.sourceStage.nextStage;
            while (sourceStage != this) {
                int sourceOrOpFlags = abstractPipeline.sourceOrOpFlags;
                if (abstractPipeline.opIsStateful()) {
                    n2 = 0;
                    if (StreamOpFlag.SHORT_CIRCUIT.isKnown(sourceOrOpFlags)) {
                        sourceOrOpFlags &= ~StreamOpFlag.IS_SHORT_CIRCUIT;
                    }
                    spliterator = abstractPipeline.opEvaluateParallelLazy(sourceStage, spliterator);
                    sourceOrOpFlags = (spliterator.hasCharacteristics(64) ? ((sourceOrOpFlags & ~StreamOpFlag.NOT_SIZED) | StreamOpFlag.IS_SIZED) : ((sourceOrOpFlags & ~StreamOpFlag.IS_SIZED) | StreamOpFlag.NOT_SIZED));
                }
                abstractPipeline.depth = n2++;
                abstractPipeline.combinedFlags = StreamOpFlag.combineOpFlags(sourceOrOpFlags, sourceStage.combinedFlags);
                sourceStage = abstractPipeline;
                abstractPipeline = abstractPipeline.nextStage;
            }
        }
        if (n != 0) {
            this.combinedFlags = StreamOpFlag.combineOpFlags(n, this.combinedFlags);
            return spliterator;
        }
        return spliterator;
    }
    
    @Override
    final StreamShape getSourceShape() {
        while (this.depth > 0) {
            this = this.previousStage;
        }
        return this.getOutputShape();
    }
    
    @Override
    final <P_IN> long exactOutputSizeIfKnown(Spliterator<P_IN> spliterator) {
        return StreamOpFlag.SIZED.isKnown(this.getStreamAndOpFlags()) ? spliterator.getExactSizeIfKnown() : -1L;
    }
    
    @Override
    final <P_IN, S extends Sink<E_OUT>> S wrapAndCopyInto(S obj, Spliterator<P_IN> spliterator) {
        this.copyInto(this.wrapSink((S)Objects.requireNonNull(obj)), spliterator);
        return obj;
    }
    
    @Override
    final <P_IN> void copyInto(Sink<P_IN> sink, Spliterator<P_IN> spliterator) {
        Objects.requireNonNull(sink);
        if (!StreamOpFlag.SHORT_CIRCUIT.isKnown(this.getStreamAndOpFlags())) {
            sink.begin(spliterator.getExactSizeIfKnown());
            spliterator.forEachRemaining(sink);
            sink.end();
        }
        else {
            this.copyIntoWithCancel(sink, spliterator);
        }
    }
    
    @Override
    final <P_IN> void copyIntoWithCancel(Sink<P_IN> sink, Spliterator<P_IN> spliterator) {
        while (this.depth > 0) {
            this = this.previousStage;
        }
        sink.begin(spliterator.getExactSizeIfKnown());
        this.forEachWithCancel(spliterator, sink);
        sink.end();
    }
    
    @Override
    final int getStreamAndOpFlags() {
        return this.combinedFlags;
    }
    
    final boolean isOrdered() {
        return StreamOpFlag.ORDERED.isKnown(this.combinedFlags);
    }
    
    @Override
    final <P_IN> Sink<P_IN> wrapSink(Sink<E_OUT> opWrapSink) {
        Objects.requireNonNull(opWrapSink);
        while (this.depth > 0) {
            opWrapSink = this.opWrapSink(this.previousStage.combinedFlags, opWrapSink);
            this = this.previousStage;
        }
        return opWrapSink;
    }
    
    @Override
    final <P_IN> Spliterator<E_OUT> wrapSpliterator(Spliterator<P_IN> spliterator) {
        if (this.depth == 0) {
            return spliterator;
        }
        return this.wrap(this, ()/*  */ -> p0, this.isParallel());
    }
    
    @Override
    final <P_IN> Node<E_OUT> evaluate(Spliterator<P_IN> spliterator, boolean b, IntFunction<E_OUT[]> intFunction) {
        if (this.isParallel()) {
            return this.evaluateToNode(this, spliterator, b, intFunction);
        }
        return ((Node.Builder<E_OUT>)this.wrapAndCopyInto(this.makeNodeBuilder(this.exactOutputSizeIfKnown(spliterator), intFunction), spliterator)).build();
    }
    
    abstract StreamShape getOutputShape();
    
    abstract <P_IN> Node<E_OUT> evaluateToNode(PipelineHelper<E_OUT> p0, Spliterator<P_IN> p1, boolean p2, IntFunction<E_OUT[]> p3);
    
    abstract <P_IN> Spliterator<E_OUT> wrap(PipelineHelper<E_OUT> p0, Supplier<Spliterator<P_IN>> p1, boolean p2);
    
    abstract Spliterator<E_OUT> lazySpliterator(Supplier<? extends Spliterator<E_OUT>> p0);
    
    abstract void forEachWithCancel(Spliterator<E_OUT> p0, Sink<E_OUT> p1);
    
    @Override
    abstract Node.Builder<E_OUT> makeNodeBuilder(long p0, IntFunction<E_OUT[]> p1);
    
    abstract boolean opIsStateful();
    
    abstract Sink<E_IN> opWrapSink(int p0, Sink<E_OUT> p1);
    
     <P_IN> Node<E_OUT> opEvaluateParallel(PipelineHelper<E_OUT> pipelineHelper, Spliterator<P_IN> spliterator, IntFunction<E_OUT[]> intFunction) {
        throw new UnsupportedOperationException("Parallel evaluation is not supported");
    }
    
     <P_IN> Spliterator<E_OUT> opEvaluateParallelLazy(PipelineHelper<E_OUT> pipelineHelper, Spliterator<P_IN> spliterator) {
        return this.opEvaluateParallel(pipelineHelper, spliterator, n/* int, */ -> (Object[])new Object[n]).spliterator();
    }
}
